#cs(module stx-xsl mzscheme
(require "sxml-tools.ss")
(require "stx-engine.ss")
(require "sxpath-ext.ss")
(require "sxpath.ss")
(require "sxpathlib.ss")
(require "libmisc.ss")
(require "fragments.ss")
(require (lib "myenv.ss" "ssax"))
(require "common.ss")
(require (lib "ssax.ss" "ssax"))
(require (lib "ssax-tools.ss" "ssax"))
(require (lib "parse-error.ss" "ssax"))
(require (lib "defmacro.ss"))
;; $Id: stx-xsl.scm,v 1.4 2004/01/31 00:38:40 kl Exp kl $

;=============================================================================
; Auxilliary

(define stx-xsl:version 
  (string-append " $Revision: 1.4 $" nl " $Date: 2004/01/31 00:38:40 $"))

; Reads content of a given SXML element 'obj' using Scheme reader.
; The content has to be a list of strings (just the first of them will be read).
; If the content is empty, "" is returned.
(define (stx:read-content obj objname)
  (let ((ct (sxml:content obj)))
    (cond 
      ((null? ct) "")
      ((string? (car ct))
	(call-with-err-handler
	  (lambda()
	    (call-with-input-string (car ct) read))
	  (lambda(mes)
	    (apply stx:error 
		      `("Error " ,nl ,mes ,nl "reading " ,objname " code:" ,nl
			,(car ct) ,nl "from element" ,nl
			,@(sxml:clean-fragments (sxml:sxml->xml obj)) ,nl))
			(exit))))
      (else 
	    (stx:error "Invalid " objname " element:" nl obj)))
      ))

; A convenience macro for stx:template-xsl->stx 
(define-macro stx:call-function 
  (lambda (name type tpl-node $-env)
  `(let ((fn (,$-env 
	      (string->symbol ,name))))
    (if 
      (eq? fn '*LT-NOT-FOUND*)
      (apply stx:error 
	     (append 
	      (list "Undefined " ,type  " with name " ,name " is called by:" nl)
	       (sxml:clean-fragments (sxml:sxml->xml ',tpl-node))
	       (list nl "Valid names: ") (map car (,$-env))
				))
      (call-with-err-handler
	(lambda()
	  (sxml:clean-fragments
 	      (fn current-node stx:templates current-root (,$-env '*LT-ADD* 
						      `(stx:param ,',tpl-node)))))
	(lambda (mes)
	  (apply stx:error 
		 (list ,type " evaluation ERROR" 
		   nl mes nl "for:" nl
		   (sxml:clean-fragments 
		     (sxml:sxml->xml ',tpl-node))))))))))

;=============================================================================
; Transformers for XSL stylesheet elements

; Transforms element stx:template (extracted from XSL stylesheet and
; represented in SXML format) to stx:template
(define (stx:scm->stx tmplt)
  `(stx:template (@ ,@(sxml:attr-list tmplt))
	(lambda (current-node stx:templates current-root $)
		 ,(stx:read-content tmplt "<stx:template@match>"))))

; Transforms STX element stx:var to stx:variable
(define (stx:stx-var->stx var)
  `(stx:variable (@ ,@(sxml:attr-list var))
		 ,(stx:read-content var "<stx:variable>")))

; Transforms XSL element xsl:var to stx:variable
(define (stx:xsl-var->stx var)
  `(stx:variable (@ ,@(sxml:attr-list var))
		 ',(sxml:content var)))

(define (stx:attr->html attr)
	 (if (equal? "" (cadr attr))
              `(list " " ,(sxml:ncname attr))
              `(list " " ,(sxml:ncname attr) "='" ,(cadr attr) "'")))

; transforms an xsl:template to stx:template
(define (stx:template-xsl->stx tmplt custom-prefixes c-root)
  (let* (
	 ; list of template's attributes
	 (attr-list (sxml:attr-list tmplt))
	 ; the name of string output function for document conversion
	 (doc-string-out 'sxml:sxml->html)
	 ; string output function for template_ conversion
	 (sst-string-out sxml:sxml->xml))

  `(stx:template (@ ,@attr-list)
	(lambda (current-node stx:templates current-root $)
	  ,(cons 'list
	  (stx:apply-templates 
	    (sxml:content tmplt)
	    `((*default* 
		 ,(lambda (tt-node bindings root environment)
		     (if (cond ((sxml:name->ns-id (car tt-node)) 
				=> (lambda(x) 
				     (member (string->symbol x)
					     custom-prefixes)))
		       (else #f))	 
		       ; user-defined tag 
		       `(stx:call-function ,(sxml:ncname tt-node) 
					   "Custom Tag"
					   ,tt-node
					   $)
		       ; XHTML tag
		       (let ((nm (sxml:ncname tt-node))
			     (content (sxml:content tt-node)))
			 (if (null? content)
			   `(list "<" ,nm ,@(map stx:attr->html 
					   (sxml:attr-list tt-node)) "/>")
			   `(list "<" ,nm ,@(map stx:attr->html 
					   (sxml:attr-list tt-node)) ">"
			     ,@(stx:apply-templates content 
						    bindings root environment)
			     "</" ,nm ">" ))))
		     ))
	      (*text*  ; text string in template 
		 ,sst-string-out)
	      (xsl:apply-templates ; t-node is <xsl:apply-templates/>
		 ,(lambda (t-node bindings root environment) 
		     `(stx:apply-templates
			,(cond 
			   ((sxml:attr t-node 'select)
			    => (lambda (lp)
				 `((sxpath ,lp) current-node 
						`((*root* ,current-root)))
				 ))
			   (else '(sxml:content current-node)))
			stx:templates current-root $)))
              (xsl:if
                ,(lambda (t-node bindings root environment)
			       ``(stx:apply-templates
                                 ,(sxml:content ,t-node)
                                 bindings
                                 root
                                 environment)
                             ))
	      (xsl:call-template ; t-node is <xsl:call-template/>
		 ,(lambda (t-node bindings root environment) 
		     `(stx:call-function ,(sxml:attr t-node 'name)
					"Named Template"
					,t-node
					$)))
	      (xsl:value-of  ; t-node is <xsl:value-of/>
		 ,(lambda (t-node bindings root environment) 
		     `(,(if (equal? "yes" 
				      (sxml:attr t-node 'disable-output-escaping))
			   'self
			   doc-string-out)
			(sxml:string 
			   ((sxpath ,(sxml:attr t-node 'select)) 
			    current-node))))) 
	      ; <xsl:copy-of/>
	      (xsl:copy-of  
		 ,(lambda (t-node bindings root environment) 
		     `((sxpath ,(sxml:attr t-node 'select)) current-node))) 
	      (stx:eval      ; t-node is <stx:eval/>
		 ,(lambda (t-node bindings root environment)  
		     (let ((content
			     (stx:read-content t-node "<stx:eval>")))
		     `(call-with-err-handler
		       (lambda()
			 (eval ,content))
		       (lambda(mes)
			 (apply stx:error `("Error " ,nl ,mes ,nl 
					    "while evaluating code:" ,nl 
					    ,,content 
					    ,nl "from element" ,nl 
					    ,@(sxml:clean-fragments 
						(sxml:sxml->xml
						  ',t-node)))))))))
	      ) c-root 
	    (lambda-tuple) ; compile-time environment
	    )))))) 

;=======================================================================
; High Level 

; A simple highest-level function:
;  transforms given XML document using supplied XSLT stylesheet
(define (stx:xsltproc doc-file stylesheet)
 (let 
  ((doc 
     (sxml:add-parents
       (ssax:xml->sxml 
	 (port-count-lines
	   (open-input-file  doc-file)) '())))
   ; The lists above as the lists of functions
   (sst-lambda (stx:eval-transformer
		 (stx:stx->tmpl+env
		    (stx:xslt->stx (ssax:xml->sxml-autoprefix stylesheet))))))
  (sxml:display-fragments
    (stx:apply-templates
      doc
      (car sst-lambda)
      doc
      (cadr sst-lambda))) 
  ))


; Generates STX stylesheet - XSL-like stylesheet programmed in Scheme -
; from an XSL stylesheet represented in SXML format
; Typically, a first step in processing XSL stylesheet
(define (stx:xslt->stx stx-tree)
  (let* 
    ((custom-prefixes 
       (map string->symbol
	    ((sxpath '((*or* xsl:stylesheet xsl:transform) 
		       stx:import @ prefix *text*))
	     stx-tree))))
     (cons 'stx:stylesheet
   (map
      (lambda(x)
	(cond 
	  ; xsl:template
	  ((eq? (car x) 'xsl:template)
	   (stx:template-xsl->stx x custom-prefixes stx-tree))
	  ; stx:template
	  ((eq? (car x) 'stx:template)
	   (stx:scm->stx x))
	  ; stx:variable
	  ((eq? (car x) 'stx:variable)
	   (stx:stx-var->stx x))
	  ; xsl:variable
	  ((eq? (car x) 'xsl:variable)
	   (stx:xsl-var->stx x))
	  (else x)))
      ((sxpath `((*or* xsl:stylesheet xsl:transform) *)) stx-tree)))
    ))


; Transform stx:stylesheet 
; to a list whose car is a list of templates (#f no templates generated) 
; and whose cadr is a list of environment variables  bindings (#f if no bindings),
; The result (so called "STX-lists stylesheet") is usually supplyed to
;  stx:eval-transformer function
(define (stx:stx->tmpl+env stx-sst)
 (let rpt ((objs ((sxpath '((*or* stx:template stx:variable)))  
		  stx-sst))
	   (templts '())
	   (envrt '()))
  (cond
    ((null? objs) (list (reverse templts)
			envrt))
    ; templates
    ((eq? (caar objs) 'stx:template)
     (let* ((obj (car objs))
	(handler (caddr obj)))
       (rpt 
	 (cdr objs)
	 (cond 
	   ((sxml:attr obj 'match) 
	    => (lambda (x) 
		 (cons 
		   (list `(sxpath ,x) handler)
		   templts)))
	   ((sxml:attr obj 'match-lambda)
	    => (lambda (x) 
		 (cons (list x handler) templts)))
	   (else 
	     (verb-2 nl "NO match for: " (cadr obj))
	     templts))
	 (cond 
	   ((sxml:attr obj 'name)
	    => (lambda (x) 
		 (cons  
		   (list (string->symbol x) handler) envrt)))
	   (else 
	     (verb-2 nl "NO name for: " (cadr obj) 
		     "==" (sxml:attr obj 'name))
	     envrt)))))
    ; variables
    ((eq? (caar objs) 'stx:variable)
      (let* ((obj (car objs))
	     (name (sxml:attr obj 'name)) 
	     (code  (caddr obj)))              ; (sxml:content obj)
	(rpt 
	(cdr objs)
	templts
	(cons (list (string->symbol name) code) envrt))))
    (else
      (verb-2 nl "Unrecognized object: " (caar objs) nl)
      (rpt (cdr objs) templts envrt)))))

; Evalutes components of given STX-lists stylesheet and returns a 
; "lambda stylesheet" where all the necessary S-expressions are evaluated
; and "default" and "text" handlers are added 
(define (stx:eval-transformer sst-scm)
  (list
      (append
	`((*default* 
	    ,(lambda (node bindings root environment)
	       (stx:apply-templates (sxml:content node) 
				    bindings 
				    root 
				    environment)))
	  (*text* ,(lambda(x) x)))
	(map 
	  (lambda(x)
	    (list (eval (car x))
		  (eval (cadr x))))
	  (car sst-scm)))
	(map 
	  (lambda(x)
	    (list (car x)
		  (eval (cadr x))))
	  (cadr sst-scm))))


(provide (all-defined)))
